/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import java.util.ArrayList;

/**
 * Array list with special print out functions.
 */
public class ExtendedList extends ArrayList {
	
	/**
	 * Retrieves an ampersand-seperated list of all string items inside this list.
	 * A new line character ('\n') is added for every three items 
	 * or when an item's length is greater than 30.
	 *
	 * @return the ampersand-seperated list.
	 */
	public String printBooleanExpression() {
		if(isEmpty()) {
			return "";
		}
		StringBuffer expr = new StringBuffer();
		
		expr.append(PDSConstant.FIVE_TABS);
		expr.append("((");
		expr.append((String) get(0));
		expr.append(")");
		for(int i = 1; i < size(); i++) {
			if (((i % 3) == 0) || (((String) get(i)).length() > 30)) {
				expr.append(PDSConstant.NEW_LINE);
				expr.append(PDSConstant.FIVE_TABS);
			}
			expr.append(" & (");
			expr.append((String) get(i));
			expr.append(")");
		}
		expr.append(")");
		
		return expr.toString();
	}
	
	public String printBoolExprRemopla() {
		if(isEmpty()) {
			return "";
		}
		StringBuffer expr = new StringBuffer();
		
		expr.append(PDSConstant.FIVE_TABS);
		expr.append("((");
		expr.append(((String) get(0)).replaceAll("=", "==").replaceAll(
			"!==", "!=").replaceAll("<==", "<=").replaceAll(">==", ">="
			).replaceAll("\\|", "\\|\\|").replaceAll("\\&", "\\&\\&"));
		expr.append(")");
		for(int i = 1; i < size(); i++) {
			if (((i % 3) == 0) || (((String) get(i)).length() > 30)) {
				expr.append(PDSConstant.NEW_LINE);
				expr.append(PDSConstant.FIVE_TABS);
			}
			expr.append(" && (");
			expr.append(((String) get(i)).replaceAll("=", "==").replaceAll(
			"!==", "!=").replaceAll("<==", "<=").replaceAll(">==", ">="
			).replaceAll("\\|", "\\|\\|").replaceAll("\\&", "\\&\\&"));
			expr.append(")");
		}
		expr.append(")");
		
		return expr.toString();
	}
	
	/**
	 * Retrieves a comma-seperated list of all string items inside this list.
	 *
	 * @return the comma-seperated list.
	 */
	public String printCommaList() {
		if(isEmpty()) {
			return "";
		}
		
		StringBuffer list = new StringBuffer();
		list.append((String) get(0));
		for(int i = 1; i < size(); i++) {
			list.append(PDSConstant.NEW_LINE);
			list.append(PDSConstant.THREE_TABS);
			list.append(PDSConstant.COMMA);
			list.append((String) get(i));
		}
		
		return list.toString();
	}
	
	/**
	 *
	 */
	public String printVarList(String type, int bits) {
		if(isEmpty()) {
			return "";
		}
		
		StringBuffer var = new StringBuffer();
		String strBit = PDSConstant.LEFT_PARENTHESIS + bits 
			+ PDSConstant.RIGHT_PARENTHESIS;
		
		for (int i = 0; i < size(); i++) {
			var.append(PDSConstant.ONE_TAB);
			var.append(type);
			var.append(PDSConstant.ONE_SPACE);
			var.append((String) get(i));
			var.append(strBit);
			var.append(PDSConstant.SEMICOLON);
			var.append(PDSConstant.NEW_LINE);
		}
		
		return var.toString();
	}
	
	/**
	 * Retrieves a comma-seperated list of fields, used in variable declarations.
	 * Object in this list is of type PDSField. The number of bits is appended
	 * each variable in parentheses.
	 *
	 * @param  bits  the number of bits of fields.
	 *
	 * @return the comma-seperated list.
	 */
	public String printFieldList(int bits) {
		if(isEmpty()) {
			return "";
		}
		
		PDSField pdsField;
		StringBuffer field = new StringBuffer();
		String strBit = PDSConstant.LEFT_PARENTHESIS + bits 
			+ PDSConstant.RIGHT_PARENTHESIS;
		String strOneBit = PDSConstant.LEFT_PARENTHESIS + "1" 
			+ PDSConstant.RIGHT_PARENTHESIS;
		
		for (int i = 0; i < size(); i++) {
			pdsField = (PDSField) get(i);
			if (pdsField.isStatic()) {
				field.append(PDSConstant.ONE_TAB);
				field.append(pdsField.getTypeVerbose());
				field.append(PDSConstant.ONE_SPACE);
				field.append(pdsField.getFieldName());
				if (pdsField.isBoolean()) {
					field.append(strOneBit);
				} else {
					field.append(strBit);
				}
				field.append(PDSConstant.SEMICOLON);
				field.append(PDSConstant.NEW_LINE);
			}
			else {
				field.append(PDSConstant.ONE_TAB);
				field.append(pdsField.getTypeVerbose());
				field.append(PDSConstant.ONE_SPACE);
				field.append(pdsField.getFieldName());
				field.append(PDSConstant.LEFT_BRACKET);
				field.append(pdsField.getFromIndex());
				field.append(PDSConstant.COMMA);
				field.append(pdsField.getToIndex());
				field.append(PDSConstant.RIGHT_BRACKET);
				if (pdsField.isBoolean()) {
					field.append(strOneBit);
				} else {
					field.append(strBit);
				}
				field.append(PDSConstant.SEMICOLON);
				field.append(PDSConstant.NEW_LINE);
			}
		}
		
		return field.toString();
	}
	
	/**
	 * Retrieves a comma-seperated list of arrays, used in variable declarations.
	 * Object in this list is of type PDSArray. The number of bits is appended
	 * each variable in parentheses.
	 *
	 * @param  bits  the number of bits of arrays.
	 *
	 * @return the comma-seperated list.
	 */
	public String printArrayList(int bits) {
		if(isEmpty()) {
			return "";
		}
		
		PDSArray pdsArray;
		StringBuffer array = new StringBuffer();
		String strBit = PDSConstant.LEFT_PARENTHESIS + bits 
			+ PDSConstant.RIGHT_PARENTHESIS;
		String strOneBit = PDSConstant.LEFT_PARENTHESIS + "1" 
			+ PDSConstant.RIGHT_PARENTHESIS;
		
		for (int i = 0; i < size(); i++) {
			pdsArray = (PDSArray) get(i);
			array.append(PDSConstant.ONE_TAB);
			array.append(pdsArray.getTypeVerbose());
			array.append(PDSConstant.ONE_SPACE);
			array.append(PDSConstant.DEFAULT_ARRAY_NAME);
			array.append(i + 1);
			array.append("[0,");
			array.append(pdsArray.getArraySize() - 1);
			array.append("]");
			if (pdsArray.isBoolean()) {
				array.append(strOneBit);
			} else {
				array.append(strBit);
			}
			array.append(PDSConstant.SEMICOLON);
			array.append(PDSConstant.NEW_LINE);
		}
		
		return array.toString();
	}
	
}
