/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.structures.*;

/**
 * Array information for the pushdown system.
 */
public class PDSArray {

	private String arrayName;
	private String className;
	private String type;
	private int arraySize;
	private int accessFlag;
	
	/**
	 * Initializes an array, access flag undefined.
	 *
	 * @param arrayName the name of this array.
	 * @param className the class this array belongs to.
	 */
	public PDSArray(String arrayName, String className) {
		this(arrayName, className, -1);
	}
	
	/**
	 * Initializes an array.
	 *
	 * @param arrayName the name of this array.
	 * @param className the class this array belongs to.
	 * @param accessFlag the access flag of this array.
	 */
	public PDSArray(String arrayName, String className, int accessFlag) {
		this.arrayName = arrayName;
		this.className = className;
		this.type = "I";
		this.accessFlag = accessFlag;
	}
	
	/**
	 * Returns the name of this array.
	 *
	 * @return array's name.
	 */
	public String getArrayName() {
		return arrayName;
	}
	
	/**
	 * Returns the name of the class that this array belongs to.
	 *
	 * @return class's name.
	 */
	public String getClassName() {
		return className;
	}
	
	/**
	 * Set the type of this array
	 *
	 * @param type the type
	 */
	public void setType(String type) {
		this.type = type;
	}
	
	/**
	 * Returns type of this array.
	 *
	 * @return array's type.
	 */
	public String getTypeVerbose() {
		
		if (!type.equals("I") && !type.equals("Z")) {
			System.err.println("Bug: unsupported type of array found");
			System.exit(1);
		}
		
		return "int";
	}
	
	/**
	 * Checks if this array is of boolean type
	 *
	 * @return <code>true</code> if boolean
	 */
	public boolean isBoolean() {
		return type.equals("Z");
	}
	
	/**
	 * Returns the size of this array.
	 *
	 * @return array's size.
	 */
	public int getArraySize() {
		return arraySize;
	}
	
	/**
	 * Set the size of this array.
	 *
	 * @param arraySize array's size.
	 */
	public void setArraySize(int arraySize) {
		this.arraySize = arraySize;
	}
	
	/**
	 * Checks whether this array is static or not.
	 *
	 * @return true, if static.
	 */
	public boolean isStatic() {
		return (accessFlag & AccessFlags.ACC_STATIC) != 0;
	}

}
