/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.structures.*;

/**
 * Basic information of a method.
 */
public class PDSBasicMethod {
	
	String className;
	String methodName;
	int accessFlag;
	boolean includedFlag;
	
	/**
	 * Initializes method name and class name.
	 *
	 * @param className the class this method belong to.
	 * @param methodName the name of this method.
	 */
	public PDSBasicMethod(String className, String methodName) {
		this.className = className;
		this.methodName = methodName;
		includedFlag = false;
	}
	
	/**
	 * Retrieves the class name of this method.
	 *
	 * @return the class name.
	 */
	public String getClassName() {
		return className;
	}
	
	/**
	 * Retrieves this method name.
	 *
	 * @return this method name.
	 */
	public String getMethodName() {
		return methodName;
	}
	
	/**
	 * Checks if this method is static.
	 *
	 * @return <code>true</code> if static.
	 */
	public boolean isStatic() {
		return (accessFlag & AccessFlags.ACC_STATIC) != 0;
	}
	
	/**
	 * Checks whether this method is already included.
	 *
	 * @return true, if already included.
	 */
	public boolean isIncluded() {
		return includedFlag;
	}
	
	/**
	 * Set the class name of this method.
	 *
	 * @param className the class name.
	 */
	public void setClassName(String className) {
		this.className = className;
	}
	
	/**
	 * Set this method name.
	 *
	 * @param methodName the method name.
	 */
	public void setMethodName(String methodName) {
		this.methodName = methodName;
	}
	
	public void  setAccessFlag(int accessFlag) {
		this.accessFlag = accessFlag;
	}
	
	/**
	 * Set its included flag.
	 *
	 * @param includedFlag the flag indicating whether this method is already included.
	 */
	public void setIncludedFlag(boolean includedFlag) {
		this.includedFlag = includedFlag;
	}
	
}
