/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import java.util.List;
import java.util.ArrayList;

/**
 * Configuration of the pushdown system.
 */
public class PDSConfig {
	
	private String controlLocation;
	private List stackSymbols;
	
	/**
	 * Initializes the configuration with the default control location.
	 *
	 * @param stackSymbol the stack alphabet.
	 */
	public PDSConfig(String stackSymbol) {
		this(PDSConstant.DEFAULT_CONTROL_LOCATION, stackSymbol);
	}
	
	/**
	 * Initializes the configuration with one stack alphabet.
	 *
	 * @param controlLocation the control location.
	 * @param stackSymbol the stack alphabet.
	 */
	public PDSConfig(String controlLocation, String stackSymbol) {
		this.controlLocation = controlLocation;
		stackSymbols = new ArrayList();
		stackSymbols.add(stackSymbol);
	}
	
	/**
	 * Initializes the configuration with two stack alphabets.
	 *
	 * @param controlLocation the control location.
	 * @param stackSymbol1 the first stack alphabet.
	 * @param stackSymbol2 the second stack alphabet.
	 */
	public PDSConfig(String controlLocation, String stackSymbol1, String stackSymbol2) {
		this.controlLocation = controlLocation;
		stackSymbols = new ArrayList();
		stackSymbols.add(stackSymbol1);
		stackSymbols.add(stackSymbol2);
	}
	
	/**
	 * Initializes the configuration with the list of stack alphabets.
	 *
	 * @param controlLocation the control location.
	 * @param stackSymbols the list of stack alphabets.
	 */
	public PDSConfig(String controlLocation, List stackSymbols) {
		this.controlLocation = controlLocation;
		this.stackSymbols = stackSymbols;
	}

	public String getFirstStackSymbol() {
		return (String) stackSymbols.get(0);
	}
	
	/**
	 * Prints the configuration.
	 *
	 * @return the configuration in Moped's format.
	 */
	public String print() {
		int stackSymbolsCount = (stackSymbols == null) ? 0 : stackSymbols.size();
		
		StringBuffer config = new StringBuffer();
		config.append(controlLocation).append(" <");
		if(stackSymbolsCount > 0)
		{
			config.append(stackSymbols.get(0));
			for(int i = 1; i < stackSymbolsCount; i++)
				config.append(" ").append(stackSymbols.get(i));
		}
		config.append(">");
		
		return config.toString();
	}
	
	/**
	 * Prints the configuration.
	 *
	 * @return the configuration in Remopla.
	 */
	public StringBuffer printRemopla() {
		StringBuffer output = new StringBuffer();
		
		output.append("init ");
		output.append(getFirstStackSymbol());
		output.append(";");
		
		return output;
	}
	
}
