/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

/**
 * Information about local variables of a method.
 */
public class PDSLocal {
	
	private ExtendedList stackSymbolList;
	private char type;
	private ExtendedList variableList;
	private int bits;
	
	/**
	 * Initializes local variables of a method
	 *
	 * @param stackSymbolList the list of stack symbols that can access this local varibles.
	 * @param type the type of local variables.
	 * @param variableList the list of local variables.
	 * @param bits the number of bits of variables.
	 */
	public PDSLocal(ExtendedList stackSymbolList, 
									char type, 
									ExtendedList variableList, 
									int bits) {
		this.stackSymbolList = stackSymbolList;
		this.type = type;
		this.variableList = variableList;
		this.bits = bits;
	}
	
	/**
	 * Returns this local variable type.
	 *
	 * @return this local variable type.
	 */
	public String getTypeVerbose() {
		if (type != 'I') {
			System.err.println("Bug: unsupported type of field found");
			System.exit(1);
		}
		
		return "int";
	}
	
	public String getFirstStackSymbol() {
		return (String) stackSymbolList.get(0);
	}
	
	/**
	 * Prints local variables declaration in Moped's syntax
	 *
	 * @return local variables declaration.
	 */
	public String print() {
		StringBuffer local = new StringBuffer();
		
		local.append("local");
		local.append(PDSConstant.ONE_SPACE);
		local.append(PDSConstant.LEFT_PARENTHESIS);
		local.append(stackSymbolList.printCommaList());
		local.append(PDSConstant.RIGHT_PARENTHESIS);
		local.append(PDSConstant.NEW_LINE);
		local.append(variableList.printVarList(getTypeVerbose(), bits));
		
		return local.toString();
	}
	
	public String printRemopla() {
		return variableList.printVarList(getTypeVerbose(), bits);
	}
	
}
