/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

/**
 * Rule of the pushdown system.
 */
public class PDSRule {
	
	private PDSConfig pdsLhsConfig;
	private PDSConfig pdsRhsConfig;
	private String pdsLabel;
	private ExtendedList pdsCond;
	
	/**
	 * Initializes a rule with empty label.
	 *
	 * @param pdsLhsConfig the left-hand-side configuration.
	 * @param pdsRhsConfig 	the right-hand-side configuration.
	 * @param pdsCond 			the condition of the rule.
	 */
	public PDSRule(PDSConfig pdsLhsConfig, 
							   PDSConfig pdsRhsConfig, 
								 ExtendedList pdsCond) {
		this(pdsLhsConfig, pdsRhsConfig, null, pdsCond);
	}
	
	/**
	 * Initializes a rule.
	 *
	 * @param pdsLhsConfig the left-hand-side configuration.
	 * @param pdsRhsConfig		the right-hand-side configuration.
	 * @param pdsLabel 			the label of the rule.
	 * @param pdsCond 			the condition of the rule.
	 */
	public PDSRule(PDSConfig pdsLhsConfig, 
							   PDSConfig pdsRhsConfig, 
								 int pdsLabel, 
								 ExtendedList pdsCond) {
		this(pdsLhsConfig, pdsRhsConfig, String.valueOf(pdsLabel), pdsCond);
	}
	
	/**
	 * Initializes a rule.
	 *
	 * @param pdsLhsConfig the left-hand-side configuration.
	 * @param pdsRhsConfig		the right-hand-side configuration.
	 * @param pdsLabel 			the label of the rule.
	 * @param pdsCond 			the condition of the rule.
	 */
	public PDSRule(PDSConfig pdsLhsConfig, 
							   PDSConfig pdsRhsConfig, 
								 String pdsLabel, 
								 ExtendedList pdsCond) {
		this.pdsLhsConfig = pdsLhsConfig;
		this.pdsRhsConfig = pdsRhsConfig;
		this.pdsLabel = pdsLabel;
		this.pdsCond = pdsCond;
	}
	
	/**
	 * Set the left-hand-side configuration of the rule.
	 *
	 * @param pdsLhsConfig the left-hand-side configuration.
	 */
	public void setPDSConfigFrom(PDSConfig pdsLhsConfig) {
		this.pdsLhsConfig = pdsLhsConfig;
	}
	
	/**
	 * Set the right-hand-side configuration of the rule.
	 *
	 * @param pdsRhsConfig the right-hand-side configuration.
	 */
	public void setPDSConfigTo(PDSConfig pdsRhsConfig) {
		this.pdsRhsConfig = pdsRhsConfig;
	}
	
	/**
	 * Set the condition of the rule.
	 *
	 * @param pdsCond the rule condition.
	 */
	public void setPDSCond(ExtendedList pdsCond) {
		this.pdsCond = pdsCond;
	}
	
	public PDSConfig getLhsConfig() {
		return pdsLhsConfig;
	}
	
	/**
	 * Prints this rule in Moped's syntax
	 *
	 * @return the string of this rule.
	 */
	public String print() {
		StringBuffer rule = new StringBuffer();
		
		rule.append(pdsLhsConfig.print());
		rule.append(PDSConstant.RULE_ARROW);
		rule.append(pdsRhsConfig.print());
		if (pdsLabel != null) {
			rule.append(PDSConstant.THREE_TABS);
			rule.append(PDSConstant.DOUBLE_QUOTE);
			rule.append(pdsLabel);
			rule.append(PDSConstant.DOUBLE_QUOTE);
		}
		rule.append(PDSConstant.NEW_LINE);
		rule.append(pdsCond.printBooleanExpression());
		
		return rule.toString();
	}
	
	public StringBuffer printRemopla() {
		StringBuffer rule = new StringBuffer();
		
		rule.append(pdsLhsConfig.print());
		rule.append(PDSConstant.RULE_ARROW);
		rule.append(pdsRhsConfig.print());
		if (pdsLabel != null) {
			rule.append(PDSConstant.THREE_TABS);
			rule.append(PDSConstant.DOUBLE_QUOTE);
			rule.append(pdsLabel);
			rule.append(PDSConstant.DOUBLE_QUOTE);
		}
		rule.append(PDSConstant.NEW_LINE);
		rule.append(pdsCond.printBoolExprRemopla());
		
		return rule;
	}
	
}
