package jmoped.cex;

import org.gjt.jclasslib.structures.*;
import org.gjt.jclasslib.bytecode.*;
import java.util.Vector;
import java.util.ArrayList;
import java.util.Stack;
import jmoped.*;

public class CounterExampleHandler {
	
	private String javaFileName;
	private ArrayList classFileList;
	private ArrayList pdsClassList;
	private Vector configList;
	
	private PDSClass pdsClass = null;
	private PDSMethod pdsMethod;
	Stack callStack = new Stack();
	
	private static final String LOCAL_VAR = "var";
	
	public CounterExampleHandler(ArrayList classFileList, 
															 ArrayList pdsClassList, 
															 String mopedOutput) throws Exception {
		this(classFileList, pdsClassList, mopedOutput, null);
	}
	
	public CounterExampleHandler(ArrayList classFileList, 
															 ArrayList pdsClassList, 
															 String mopedOutput, 
															 String javaFileName) throws Exception {
		this.classFileList = classFileList;
		this.pdsClassList = pdsClassList;
		this.javaFileName = javaFileName;
		configList = CounterExampleParser.parse(mopedOutput);
		if (PDSConstant.verbose > 1) {
			System.out.println(print(configList).toString());
		}
	}
	
	public void interpret() throws Exception {
		Vector config = (Vector) configList.elementAt(0);
		Vector next_config;
		Vector stack;
		Vector stackElement;
		String stackSymbol;
		String methodName;
		String formattedMethodName;
		StringBuffer methodDescriptor;
		String startStackSymbol;
		int configIndex;
		MethodInfo[] methodInfo;
		Vector label;
		int opcode;
		int storeIndex;
		StringBuffer cex = new StringBuffer();
		StringBuffer stmt;
		int paramCount;
		String compOp = null;
		
		/* Starts main method or not. */
		pdsClass = (PDSClass) pdsClassList.get(0);
		formattedMethodName = pdsClass.getClassName() 
			+ "_main_Ljava_lang_String_V";
		pdsMethod = pdsClass.getPDSMethod(formattedMethodName);
		startStackSymbol = formattedMethodName + "0";
		for (configIndex = 0; configIndex < configList.size(); configIndex++) {
			config = (Vector) configList.elementAt(configIndex);
			stack = (Vector) config.elementAt(3);
			stackElement = (Vector) stack.elementAt(0);
			stackSymbol = (String) stackElement.elementAt(0);
			if (stackSymbol.equals(startStackSymbol))
				break;
		}
		callStack.push(pdsMethod);
		
		/* Reports error if main method not found */
		if (configIndex == configList.size()) {
			System.err.println("Bug: Witness path does not contain main method");
			System.exit(1);
		}
		
		for (; configIndex < configList.size(); configIndex++) {
			config = (Vector) configList.elementAt(configIndex);
			
			if (config.size() > 4) {
				cex.append("--- LOOP ---\n");
			}
			
			label = (Vector) config.elementAt(0);
			if (label == null) {
				continue;
			}
			opcode = Integer.parseInt((String) label.elementAt(0));
			storeIndex = -1;
			stmt = new StringBuffer();
			if (PDSConstant.verbose > 2) {
				System.out.println(printLabel(label));
			}
			
			switch (opcode) {
				
				case Opcodes.OPCODE_ASTORE:
					storeIndex = Integer.parseInt((String) label.elementAt(1));
				case Opcodes.OPCODE_ASTORE_0:
					if (storeIndex == -1) storeIndex = 0;
				case Opcodes.OPCODE_ASTORE_1:
					if (storeIndex == -1) storeIndex = 1;
				case Opcodes.OPCODE_ASTORE_2:
					if (storeIndex == -1) storeIndex = 2;
				case Opcodes.OPCODE_ASTORE_3:
					if (storeIndex == -1) storeIndex = 3;
					stmt.append(printTab(callStack.size()));
					stmt.append(LOCAL_VAR);
					stmt.append(storeIndex);
					stmt.append(" = ");
					stmt.append(buildExpr(configIndex - 1));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
					
				case Opcodes.OPCODE_IASTORE: {
					int index = configIndex - 1;
					
					stmt.append(buildExpr(index));
					index -= countConfig(index);
					stmt.insert(0, "] = ");
					stmt.insert(0, buildExpr(index, false));
					stmt.insert(0, "[");
					index -= countConfig(index);
					stmt.insert(0, buildExpr(index, false));
					stmt.insert(0, printTab(callStack.size()));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
				}
				
				case Opcodes.OPCODE_IINC:
					stmt.append(buildInc(label));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
				
				case Opcodes.OPCODE_INVOKESPECIAL: {
					String calledMethod = (String) label.elementAt(2);
					
					if (calledMethod.equals("java/lang/AssertionError.<init>")) {
						stmt.append(printTab(callStack.size()));
						stmt.append("AssertionError();\n");
						break;
					}
					if (calledMethod.equals("java/lang/Object.<init>")
						|| calledMethod.equals("java/lang/StringBuffer.<init>")) {
						break;
					}
					stmt.append(buildInvoke(configIndex - 1, label, false));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
				}
					
				case Opcodes.OPCODE_INVOKESTATIC: {
					String calledMethod = (String) label.elementAt(2);
					
					if (calledMethod.equals("java/lang/Integer.parseInt")) {
						break;
					}
					stmt.append(buildInvoke(configIndex - 1, label, true));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
				}
					
				case Opcodes.OPCODE_INVOKEVIRTUAL: {
					String calledMethod = (String) label.elementAt(2);
				
					if (calledMethod.equals("java/lang/StringBuffer.append")
						|| calledMethod.equals("java/lang/StringBuffer.toString")
						|| calledMethod.equals("java/io/PrintStream.print")
						|| calledMethod.equals("java/io/PrintStream.println")) {
						break;
					}
					stmt.append(buildInvoke(configIndex - 1, label, false));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
				}
					
				case Opcodes.OPCODE_ISTORE:
					storeIndex = Integer.parseInt((String) label.elementAt(1));
				case Opcodes.OPCODE_ISTORE_0:
					if (storeIndex == -1) storeIndex = 0;
				case Opcodes.OPCODE_ISTORE_1:
					if (storeIndex == -1) storeIndex = 1;
				case Opcodes.OPCODE_ISTORE_2:
					if (storeIndex == -1) storeIndex = 2;
				case Opcodes.OPCODE_ISTORE_3:
					if (storeIndex == -1) storeIndex = 3;
					stmt.append(printTab(callStack.size()));
					stmt.append(LOCAL_VAR);
					stmt.append(storeIndex);
					stmt.append(" = ");
					stmt.append(buildExpr(configIndex - 1));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
					
				case Opcodes.OPCODE_PUTFIELD: {
					int index = configIndex - 1;
					String field = (String) label.elementAt(2);
					StringBuffer var;
				
					stmt.append(buildExpr(index));
					stmt.insert(0, " = ");
					index = index - countConfig(index);
					var = buildExpr(index);
					stmt.insert(0, field.substring(field.lastIndexOf(".")));
					stmt.insert(0, var);
					stmt.insert(0, printTab(callStack.size()));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
				}
					
				case Opcodes.OPCODE_PUTSTATIC:
					stmt.append(printTab(callStack.size()));
					stmt.append((String) label.elementAt(2));
					stmt.append(" = ");
					stmt.append(buildExpr(configIndex - 1));
					stmt.append(printStmtEnding(stmt, configIndex));
					break;
				
				case Opcodes.OPCODE_RETURN:
					stmt.append(printTab(callStack.size()));
					stmt.append("return;\n");
					callStack.pop();
					pdsMethod = (PDSMethod) callStack.peek();
					break;
				
				case Opcodes.OPCODE_IRETURN:
					stmt.append(printTab(callStack.size()));
					stmt.append("return ");
					stmt.append(buildExpr(configIndex - 1));
					stmt.append(";\n");
					callStack.pop();
					pdsMethod = (PDSMethod) callStack.peek();
					break;
					
				case Opcodes.OPCODE_IFEQ:
					compOp = " == 0";
				case Opcodes.OPCODE_IFNE: {
					if (compOp == null) compOp = " != 0";
					Vector preConfig = (Vector) configList.elementAt(configIndex - 1);
					Vector preLabel = (Vector) preConfig.elementAt(0);
					int preOpcode = Integer.parseInt((String) preLabel.elementAt(0));
					if (preOpcode == Opcodes.OPCODE_GETSTATIC) {
						if (((String) preLabel.elementAt(2)).endsWith(
								PDSConstant.ASSERTIONS_DISABLED)) {
							compOp = null;
							break;
						}
					}
				}
				case Opcodes.OPCODE_IFLT:
					if (compOp == null) compOp = " < 0";
				case Opcodes.OPCODE_IFGE:
					if (compOp == null) compOp = " >= 0";
				case Opcodes.OPCODE_IFGT:
					if (compOp == null) compOp = " > 0";
				case Opcodes.OPCODE_IFLE:
					if (compOp == null) compOp = " <= 0";
				case Opcodes.OPCODE_IFNULL:
					if (compOp == null) compOp = " == null";
				case Opcodes.OPCODE_IFNONNULL:
					if (compOp == null) compOp = " != null";
					stmt.append(buildIf(callStack.size(), configIndex, compOp));
					stmt.append(printStmtEnding(stmt, configIndex));
					compOp = null;
					break;
					
				case Opcodes.OPCODE_IF_ICMPEQ:
				case Opcodes.OPCODE_IF_ACMPEQ: 
					if (compOp == null) compOp = " == ";
				case Opcodes.OPCODE_IF_ICMPNE:
				case Opcodes.OPCODE_IF_ACMPNE: 
					if (compOp == null) compOp = " != ";
				case Opcodes.OPCODE_IF_ICMPLT:
					if (compOp == null) compOp = " < ";
				case Opcodes.OPCODE_IF_ICMPGE:
					if (compOp == null) compOp = " >= ";
				case Opcodes.OPCODE_IF_ICMPGT:
					if (compOp == null) compOp = " > ";
				case Opcodes.OPCODE_IF_ICMPLE:
					if (compOp == null) compOp = " <= ";
					stmt.append(buildIfcmp(callStack.size(), configIndex, compOp));
					stmt.append(printStmtEnding(stmt, configIndex));
					compOp = null;
					break;
			}
			if (PDSConstant.verbose > 2) {
				System.out.println(stmt);
			}
			cex.append(stmt);
		}
		
		System.out.println("");
		System.out.println("Witness Path:");
		System.out.println(cex);
	}
	
	private StringBuffer buildInc(Vector label) {
		StringBuffer stmt = new StringBuffer();
		int incBy = Integer.parseInt((String) label.elementAt(2));
					
		stmt.append(printTab(callStack.size()));
		stmt.append(LOCAL_VAR);
		stmt.append((String) label.elementAt(1));
		if (incBy == 1) {
			stmt.append("++");
		} else if (incBy == 255) {
			stmt.append("--");
		} else if (incBy < 128){
			stmt.append(" += ");
			stmt.append(incBy);
		} else {
			stmt.append(" += ");
			stmt.append(incBy - 256);
		}
		
		return stmt;
	}
	
	private StringBuffer buildInvoke(int index, Vector label, boolean sttc) {
		StringBuffer stmt = new StringBuffer();
		String formattedMethodName;
		String methodName;
		
		if (label.size() < 3) {
			return stmt;
		}
		
		formattedMethodName = PDSClass.getPDSMethodSignature(
			((String) label.elementAt(2)) 
			+ printMethodDescriptor(label).toString());
		pdsMethod = findPDSMethod(formattedMethodName);
		
		if (pdsMethod.getParamCount() > 0) {
			for (int i = 0; i < (pdsMethod.getParamCount() - 1); i++) {
				stmt.insert(0, buildExpr(index));
				stmt.insert(0, ",");
				index = index - countConfig(index);
			}
			stmt.insert(0, buildExpr(index));
		}
		stmt.insert(0, "(");
		methodName = (String) label.elementAt(2);
		if (methodName.endsWith(".<init>")) {
			methodName = methodName.substring(0, methodName.indexOf(".<init>"));
		} else if (!sttc) {
			index = index - countConfig(index);
			methodName = buildExpr(index).append(methodName.substring(
				methodName.lastIndexOf("."))).toString();
		}
		stmt.insert(0, methodName);
		stmt.insert(0, printTab(callStack.size()));
		stmt.append(")");
		callStack.push(pdsMethod);
		
		return stmt;
	}
	
	private StringBuffer buildIf(int tab, 
															 int configIndex, 
															 String compOp) {
		StringBuffer stmt = new StringBuffer();
																 
		stmt.append(printTab(tab));
		stmt.append("if (");
		stmt.append(buildExpr(configIndex - 1));
		stmt.append(compOp);
		stmt.append(")");
		
		return stmt;
	}
	
	private StringBuffer buildIfcmp(int tab, 
															 		int configIndex, 
																	String compOp) {
		StringBuffer stmt = new StringBuffer();
		
		stmt.append(compOp);
		stmt.append(buildExpr(configIndex - 1));
		stmt.insert(0, 
			buildExpr(configIndex - countConfig(configIndex - 1) - 1));
		stmt.insert(0, "if (");
		stmt.insert(0, printTab(tab));
		stmt.append(")");
		
		return stmt;
	}
	
	private StringBuffer printStmtEnding(StringBuffer stmt, 
																		int configIndex) {
		StringBuffer ending = new StringBuffer();
		int skip;

		ending.append(";");
		ending.append(printSpace(stmt));
		if (configIndex + 1 < configList.size()) {
			skip = (pdsMethod.isStatic()) ? pdsMethod.getMaxStack()
																		: pdsMethod.getMaxStack() + 1;
			ending.append(printCurrentValue(configIndex, skip));
		}
		ending.append("\n");
		
		return ending;
	}
	
	private StringBuffer printCurrentValue(int configIndex, 
																				 int skip) {
		Vector config = (Vector) configList.elementAt(configIndex + 1);
		Vector stack = (Vector) config.elementAt(3);
		Vector stackElement = (Vector) stack.elementAt(0);
		Vector exprList;
		StringBuffer value = new StringBuffer();
		
		exprList = (Vector) config.elementAt(2);				// Global
		value.append(buildVar(exprList, 0, true));
		exprList = (Vector) stackElement.elementAt(1);	// Local.
		value.append(buildVar(exprList, skip, false));
		
		return value;
	}
	
	private StringBuffer buildVar(Vector exprList, 
																int start, 
																boolean global) {
		StringBuffer var = new StringBuffer();
		Vector expr;
		String varName;
		PDSGlobal pdsGlobal = pdsClass.getPDSGlobal();
		
		int count = 0;
		
		if ((exprList == null) || (start >= exprList.size())) {
			return var;
		}
		
		/* Rename array variable */
		if (global && ((count = pdsGlobal.getArrayCount()) > 0)) {
			String arrayName;
			PDSArray pdsArray;
			
			for (int i = 0; i < count; i++) {
				pdsArray = pdsGlobal.getPDSArray(i);
				arrayName = PDSConstant.DEFAULT_ARRAY_NAME + String.valueOf(i + 1);
				for (int j = 0; j < exprList.size(); j++) {
					expr = (Vector) exprList.elementAt(j);
					varName = (String) expr.elementAt(0);
					if (varName.equals(arrayName)) {
						expr.setElementAt(pdsArray.getArrayName(), 0);
						exprList.setElementAt(expr, j);
					}
				}
				start++;
			}
		}
		
		if (!global) {
			var.append("\n                    ");
			var.append("                    ");
		}
		
		/* For each variable in exprList */
		for (int i = start; i < exprList.size(); i++) {
			if (!global && pdsMethod.isParamArray(i - start)) {
				continue;
			}
			
			expr = (Vector) exprList.elementAt(i);
			varName = (String) expr.elementAt(0);
			if (varName.equals("ret")) {
				continue;
			}
			
			if (global) {
				var.append("\n                    ");
				var.append("                    ");
			}
			var.append(varName.replaceAll("_","."));
			var.append("=");
			if (expr.size() > 2) {
				var.append("{ ");
				while (varName.equals((String) expr.elementAt(0))) {
					var.append((String) expr.elementAt(1));
					var.append(" ");
					i++;
					if (i >= exprList.size()) {
						break;
					}
					expr = (Vector) exprList.elementAt(i);
				}
				i--;
				var.append("}");
			} else {
				var.append((String) expr.elementAt(1));
				var.append(" ");
			}
		}
		
		return var;
	}
	
	private StringBuffer buildExpr(int configIndex) {
		return buildExpr(configIndex, true);
	}
	
	private StringBuffer buildExpr(int configIndex, 
																 boolean outmost) {
		Vector config;
		Vector label;
		int opcode;
		StringBuffer expr = new StringBuffer();
		String binaryOp = null;
		
		config = (Vector) configList.elementAt(configIndex);
		label = (Vector) config.elementAt(0);
		opcode = Integer.parseInt((String) label.elementAt(0));
		
		switch (opcode) {
			case Opcodes.OPCODE_ACONST_NULL:
				expr.append("null");
				break;
				
			case Opcodes.OPCODE_ALOAD:
				expr.append(LOCAL_VAR);
				expr.append((String) label.elementAt(1));
				break;
				
			case Opcodes.OPCODE_ALOAD_0:
			case Opcodes.OPCODE_ALOAD_1:
			case Opcodes.OPCODE_ALOAD_2:
			case Opcodes.OPCODE_ALOAD_3:
				expr.append(LOCAL_VAR);
				expr.append(opcode - Opcodes.OPCODE_ALOAD_0);
				break;
			
			case Opcodes.OPCODE_ARRAYLENGTH:
				expr.append(buildExpr(configIndex - 1));
				expr.append(".length");
				break;
			
			case Opcodes.OPCODE_BIPUSH:
			case Opcodes.OPCODE_SIPUSH:
				expr.append((String) label.elementAt(1));
				break;
				
			case Opcodes.OPCODE_ICONST_0:
			case Opcodes.OPCODE_ICONST_1:
			case Opcodes.OPCODE_ICONST_2:
			case Opcodes.OPCODE_ICONST_3:
			case Opcodes.OPCODE_ICONST_4:
			case Opcodes.OPCODE_ICONST_5:
				expr.append(opcode - Opcodes.OPCODE_ICONST_0); 
				break;
				
			case Opcodes.OPCODE_ICONST_M1:
				expr.append("-1");
				break;
				
			case Opcodes.OPCODE_ILOAD:
				expr.append(LOCAL_VAR);
				expr.append((String) label.elementAt(1));
				break;
				
			case Opcodes.OPCODE_ILOAD_0:
			case Opcodes.OPCODE_ILOAD_1:
			case Opcodes.OPCODE_ILOAD_2:
			case Opcodes.OPCODE_ILOAD_3:
				expr.append(LOCAL_VAR);
				expr.append(opcode - Opcodes.OPCODE_ILOAD_0);
				break;
				
			case Opcodes.OPCODE_IALOAD:
				expr.append(buildExpr(configIndex - 1, false));
				expr.insert(0, "[");
				expr.insert(0, buildExpr(
					configIndex - countConfig(configIndex - 1) - 1, false));
				expr.append("]");
				break;
				
			case Opcodes.OPCODE_GETSTATIC:
				expr.append((String) label.elementAt(2));
				break;
				
			case Opcodes.OPCODE_GETFIELD: {
				String field = (String) label.elementAt(2);
				StringBuffer var = buildExpr(configIndex - 1);
				
				expr.append(var);
				expr.append(field.substring(field.lastIndexOf(".")));
				break;
			}
				
			case Opcodes.OPCODE_IADD:
				if (binaryOp == null) binaryOp = "+";
			case Opcodes.OPCODE_ISUB:
				if (binaryOp == null) binaryOp = "-";
			case Opcodes.OPCODE_IMUL:
				if (binaryOp == null) binaryOp = "*";
			case Opcodes.OPCODE_IDIV:
				if (binaryOp == null) binaryOp = "/";
			case Opcodes.OPCODE_ISHL:
				if (binaryOp == null) binaryOp = "<<";
				expr.append(" ");
				expr.append(binaryOp);
				expr.append(" ");
				expr.append(buildExpr(configIndex - 1, false));
				if (!outmost) expr.append(")");
				expr.insert(0, buildExpr(
					configIndex - countConfig(configIndex - 1) - 1, false));
				if (!outmost) expr.insert(0, "(");
				break;
				
			case Opcodes.OPCODE_INEG:
				expr.append("-");
				expr.append(buildExpr(configIndex - 1));
				
			case Opcodes.OPCODE_IRETURN:	{	// Last instruction return something.
				int index = configIndex;
				Vector preConfig;
				Vector preLabel;
				int preOpcode;
				
				do {
					index--;
					preConfig = (Vector) configList.elementAt(index);
					preLabel = (Vector) preConfig.elementAt(0);
					preOpcode = Integer.parseInt((String) preLabel.elementAt(0));
					//System.out.println("  " + printLabel(preLabel));
				} while (!isInvokeOpcode(preOpcode));
				
				if (preLabel.size() > 4) {
					int paramCount = preLabel.size() - 4;
					for (int i = 0; i < (paramCount - 1); i++) {
						expr.insert(0, buildExpr(index - 1));
						expr.insert(0, ",");
						index = index - countConfig(index - 1) -1;
					}
					expr.insert(0, buildExpr(index - 1));
				}
				expr.insert(0, "(");
				expr.insert(0, (String) preLabel.elementAt(2));
				expr.append(")");
				break;
			}
			
			case Opcodes.OPCODE_INVOKESTATIC:
				expr.append("<Nondeterministic>");
				break;
			
			/* This can happen only in case of invokespecial <init> */
			case Opcodes.OPCODE_RETURN: {
				int index = configIndex;
				Vector preConfig;
				Vector preLabel;
				int preOpcode;
				String calledMethod = "";
				
				do {
					index--;
					preConfig = (Vector) configList.elementAt(index);
					preLabel = (Vector) preConfig.elementAt(0);
					preOpcode = Integer.parseInt((String) preLabel.elementAt(0));
					if (preLabel.size() > 2) {
						calledMethod = (String) preLabel.elementAt(2);
					}
					//System.out.println("  " + printLabel(preLabel));
				} while ((preOpcode != Opcodes.OPCODE_INVOKESPECIAL) ||
						(calledMethod.equals("java/lang/Object.<init>")));
				
				expr.append("new ");
				expr.append(calledMethod.substring(0, calledMethod.length() - 7));
				expr.append("(");
				if (preLabel.size() > 4) {
					int paramCount = preLabel.size() - 4;
					for (int i = 0; i < (paramCount - 1); i++) {
						expr.append(buildExpr(index - 1));
						index = index - countConfig(index - 1) -1;
					}
					expr.append(buildExpr(index - 1));
				}
				expr.append(")");
				break;
			}
			
			default:
				expr.append(buildExpr(configIndex - 1, false));
				break;
		}
		
		return expr;
	}
	
	/**
	 * Counts the number of configurations involved in pushing 1
	 * element into variable stack.
	 */
	private int countConfig(int configIndex) {
		Vector config;
		Vector label;
		int opcode;
		int depth = 1;
		int i;
		
		for (i = configIndex; i >= 0; i--) {
			config = (Vector) configList.elementAt(i);
			label = (Vector) config.elementAt(0);
			if (PDSConstant.verbose > 2) {
				System.out.println("  " + printLabel(label) + "  " + depth);
			}
			opcode = Integer.parseInt((String) label.elementAt(0));
			if (isPushOpcode(opcode)) {
				depth--;
			} else if (isBinaryOpcode(opcode) || isPopOpcode(opcode)) {
				depth++;
			} else if (isPopPopOpcode(opcode)) {
				depth += 2;
			} else if (isNoPushPopOpcode(opcode)) {
				;
			} else if (isInvokeOpcode(opcode)) {
				int paramCount = label.size() - 4;
				int retCount = label.lastElement().equals("V") ? 0 : 1;
				if (opcode != Opcodes.OPCODE_INVOKESTATIC) {
					paramCount++;
				}
				depth = depth + paramCount - retCount;
			} else {
				System.err.println("Bug: Unknown opcode (" + opcode
					+ ") found at countConfig");
				System.exit(1);
			}
			
			if (depth == 0) {
				break;
			}
		}
		
		return configIndex - i + 1;
	}
	
	private boolean isPushOpcode(int opcode) {
		switch (opcode) {
			case Opcodes.OPCODE_ACONST_NULL:
			case Opcodes.OPCODE_ALOAD:
			case Opcodes.OPCODE_ALOAD_0:
			case Opcodes.OPCODE_ALOAD_1:
			case Opcodes.OPCODE_ALOAD_2:
			case Opcodes.OPCODE_ALOAD_3:
			case Opcodes.OPCODE_BIPUSH:
			case Opcodes.OPCODE_ICONST_0:
			case Opcodes.OPCODE_ICONST_1:
			case Opcodes.OPCODE_ICONST_2:
			case Opcodes.OPCODE_ICONST_3:
			case Opcodes.OPCODE_ICONST_4:
			case Opcodes.OPCODE_ICONST_5:
			case Opcodes.OPCODE_ICONST_M1:
			case Opcodes.OPCODE_ILOAD:
			case Opcodes.OPCODE_ILOAD_0:
			case Opcodes.OPCODE_ILOAD_1:
			case Opcodes.OPCODE_ILOAD_2:
			case Opcodes.OPCODE_ILOAD_3:
			case Opcodes.OPCODE_GETSTATIC:
			case Opcodes.OPCODE_NEW:
			case Opcodes.OPCODE_DUP:
				return true;
		}
		return false;
	}
	
	private boolean isPopOpcode(int opcode) {
		switch (opcode) {
			case Opcodes.OPCODE_ISTORE:
			case Opcodes.OPCODE_ISTORE_0:
			case Opcodes.OPCODE_ISTORE_1:
			case Opcodes.OPCODE_ISTORE_2:
			case Opcodes.OPCODE_ISTORE_3:
			case Opcodes.OPCODE_PUTSTATIC:
			
			case Opcodes.OPCODE_IALOAD:
			case Opcodes.OPCODE_AALOAD:
				return true;
		}
		return false;
	}
	
	private boolean isBinaryOpcode(int opcode) {
		switch (opcode) {
			case Opcodes.OPCODE_IADD:
			case Opcodes.OPCODE_ISUB:
			case Opcodes.OPCODE_IMUL:
			case Opcodes.OPCODE_IDIV:
			case Opcodes.OPCODE_ISHL:
				return true;
		}
		return false;
	}
	
	private boolean isPopPopOpcode(int opcode) {
		switch (opcode) {
			case Opcodes.OPCODE_PUTFIELD:
				return true;
		}
		return false;
	}
	
	private boolean isNoPushPopOpcode(int opcode) {
		switch (opcode) {
			case Opcodes.OPCODE_ARRAYLENGTH:
			case Opcodes.OPCODE_RETURN:
			case Opcodes.OPCODE_GETFIELD:
				return true;
		}
		return false;
	}
	
	private boolean isInvokeOpcode(int opcode) {
		switch (opcode) {
			case Opcodes.OPCODE_INVOKEVIRTUAL:
			case Opcodes.OPCODE_INVOKESPECIAL:
			case Opcodes.OPCODE_INVOKESTATIC:
			case Opcodes.OPCODE_INVOKEINTERFACE:
				return true;
		}
		return false;
	}
	
	public static StringBuffer printExprList(Vector exprList) {
		StringBuffer out = new StringBuffer();
		Vector expr;
		
		out.append(" ( ");
		for (int i = 0; i < exprList.size(); i++) {
			expr = (Vector) exprList.elementAt(i);
			out.append((String) expr.elementAt(0));
			if (expr.size() > 2) {
				out.append("[");
				out.append((String) expr.elementAt(2));
				out.append("]");
			}
			out.append("=");
			out.append((String) expr.elementAt(1));
			out.append(" ");
		}
		out.append(")");
		
		return out;
	}
	
	public static StringBuffer printStack(Vector stack) {
		StringBuffer out = new StringBuffer();
		Vector stackElement;
		
		for (int i =0; i < stack.size(); i++) {
			stackElement = (Vector) stack.elementAt(i);
			out.append((String) stackElement.elementAt(0));
			out.append(" ");
			if (stackElement.elementAt(1) != null) {
				out.append(printExprList((Vector) stackElement.elementAt(1)));
			}
			out.append("\n");
		}
		
		return out;
	}
	
	public static StringBuffer printLabel(Vector label) {
		StringBuffer out = new StringBuffer();
		
		out.append("[");
		for (int j = 0; j < label.size(); j++) {
			out.append(" ");
			out.append((String) label.elementAt(j));
		}
		out.append(" ]");
		
		return out;
	}
	
	public static StringBuffer print(Vector configList) {
		StringBuffer out = new StringBuffer();
		Vector config;
		Vector label;
		
		for (int i = 0; i < configList.size(); i++) {
			config = (Vector) configList.elementAt(i);
			
			/* Label */
			if (config.elementAt(0) != null) {
				out.append(printLabel((Vector) config.elementAt(0)));
				out.append("\n");
			}
			
			/* Control Location */
			out.append((String) config.elementAt(1));
			
			/* Expression List */
			if (config.elementAt(2) != null) {
				out.append(printExprList((Vector) config.elementAt(2)));
			}
			out.append("\n<");
			
			/* Stack */
			if (config.elementAt(3) != null) {
				out.append(printStack((Vector) config.elementAt(3)));
			}
			out.append(">\n");
			
		}
		
		return out;
	}
	
	private PDSMethod findPDSMethod(String methodName) {
		PDSClass tmpClass;
		PDSMethod tmpMethod = null;
		
		for (int i = 0; i < pdsClassList.size(); i++) {
			tmpClass = (PDSClass) pdsClassList.get(i);
			tmpMethod = tmpClass.getPDSMethod(methodName);
			if (tmpMethod != null) {
				break;
			}
		}
		
		if (tmpMethod == null) {
			System.err.println("Bug: Cannot find method " + methodName
				+ " in CounterExampleHandler.findPDSMethod()");
			System.exit(1);
		}
		
		return tmpMethod;
	}
	
	private StringBuffer printMethodDescriptor(Vector label) {
		StringBuffer methodDescriptor = new StringBuffer();
		String param;
		
		methodDescriptor.append("(");
		for (int i = 3; i < label.size() - 1; i++) {
			param = (String) label.elementAt(i);
			methodDescriptor.append(param);
			if (param.charAt(0) == 'L') {
				methodDescriptor.append(";");
			}
		}
		methodDescriptor.append(")");
		methodDescriptor.append((String) label.lastElement());
		
		return methodDescriptor;
	}
	
	private StringBuffer printTab(int quan) {
		StringBuffer tab = new StringBuffer();
		
		for (int i = 0; i < quan; i++) {
			tab.append("  ");
		}
		
		return tab;
	}
	
	private StringBuffer printSpace(StringBuffer in) {
		StringBuffer out = new StringBuffer(" ");
		
		for (int i = in.length(); i < 40; i++) {
			out.append(" ");
		}
		
		return out;
	}
}
