/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import java.io.*;
import jmoped.cex.*;

class jmoped {
	static void usage(String name) {
		System.err.println(
			"\n" + name + " -- java model checker\n\n" +
			
			"Usage: " + name + " [-<options>] <classfile>\n\n" +
			
			"    Options:\n" +
			"    -a\t\t check violation of assertion\n"+
			"    -<n>\t set number of bits of integer variables (default=4)\n" +
			"    -Dx=n\t define number of bits of a scalar variable\n" +
			"    -e\t\t check reachability of error method\n" +
			"    -n\t\t output new syntax" +
			"    -o<..>\t set output file name\n" +
			"    -s<n>\t set verbosity\n" +
			"    -t\t\t check termination\n" +
			"    -w\t\t print counterexample\n\n" +
			
			"Copyright (C) 2004-2005 \n" + 
			"Dejvuth Suwimonteerabuth " +
			"<suwimodh@fmi.uni-stuttgart.de>\n\n" +
			"See COPYING for legal information.\n\n" +
			
			"Version 1.0.1 (11.04.2005)\n"
			);
		
		System.exit(1);
	}
	
	static void optError() {
		System.err.println("Options -a, -e and -t cannot mix together\n");
		System.exit(1);
	}
	
	public static void main(String[] args) throws Exception {
		
		String toolName = "jmoped";
		String inFileName = "";
		String outFileName = "";
		String classFileName;
		String ltlFileName = "__temp.ltl";
		String className;
		String mopedForm = "";
		int bits = 4;
		StringBuffer dOpt = new StringBuffer();
		PDSConstant.verbose = 0;
		boolean chkAssert = false;
		boolean chkErr = false;
		boolean chkTerm = false;
		boolean chkLTL = false;
		boolean chkReach = false;
		boolean javaInput = false;
		boolean newSyntax = false;
		boolean printCex = false;
		StringBuffer mopedOpt = new StringBuffer();
		File ltlFile = null;
		
		if (args.length == 0) {
			usage(toolName);
		}
		
		for (int i = 0; i < args.length; i++) {
			if (args[i].startsWith("-")) {
				for (int j = 1; j < args[i].length(); j++) {
					if (Character.isDigit(args[i].charAt(j))) {
						if ((j < args[i].length() - 1) 
								&& Character.isDigit(args[i].charAt(j+1))) {
							bits = Integer.parseInt(args[i].substring(j, j+2));
							j++;
						}
						else
							bits = Character.getNumericValue(args[i].charAt(j));
					} else {
						switch (args[i].charAt(j)) {
							
							case 'a': 
								if (chkErr || chkTerm || chkLTL || chkReach) {
									optError();
								}
								chkAssert = true;
								break;
							
							case 'D': 
								dOpt.append(args[i].substring(j+1)).append(" ");
								j = args[i].length();
								break;
							
							case 'e': 
								if (chkAssert || chkLTL || chkReach || chkTerm) {
									optError();
								}
								chkErr = true;
								break;
							
							case 'l': 
								if (chkAssert || chkErr || chkReach || chkTerm) {
									optError();
								}
								chkLTL = true;
								break;
								  
							case 'M': 
								mopedOpt.append(args[i].substring(j+1));
								j = args[i].length();
								break;
								
							case 'n': 
								newSyntax = true;
								break;
							
							case 'o': 
								outFileName = args[i].substring(j+1);
								j = args[i].length();
								break;
							
							case 'r': 
								if (chkAssert || chkErr || chkLTL || chkTerm) {
									optError();
								}
								chkReach = true;
								break;
							
							case 's': 
								if (Character.isDigit(args[i].charAt(++j))) {
									PDSConstant.verbose = Character.getNumericValue(
										args[i].charAt(j));
								} else {
									usage(toolName);
								}
								break;
							
							case 't': 
								if (chkAssert || chkErr || chkLTL || chkReach) {
									optError();
								}
								chkTerm = true;
								break;
								  
							case 'w': 
								printCex = true;
								mopedOpt.append("t");
								break;
							
							default: 
								usage(toolName);
								break;
						}
					}
				}
			} else {
				if (inFileName.length() == 0) {
					inFileName = args[i];
				} else if ((chkLTL || chkReach) && (mopedForm.length() == 0)) {
					mopedForm = args[i];
				} else {
					usage(toolName);
				}
			}
		}
		
		//Formats input file name
		if (inFileName.length() == 0) {
			usage(toolName);
		}
		
		//If input is Java source, compiles it.
		if (inFileName.endsWith(".java")) {
			javaInput = true;
			classFileName = inFileName.substring(0,inFileName.length() - 5)
				+ ".class";
			com.sun.tools.javac.Main javac = new com.sun.tools.javac.Main();
			String[] javacArgs = new String[] {"-source", "1.4", inFileName};
			int status = javac.compile(javacArgs);
			if (status != 0) {
				System.exit(1);
			}
		} else if (!inFileName.endsWith(".class")) {
			classFileName = inFileName.concat(".class");
		} else {
			classFileName = inFileName;
		}
		
		//Gets class name (without ".class")
		int sepIndex = classFileName.lastIndexOf(File.separatorChar);
		if (sepIndex == -1) {
			className = classFileName.substring(0, 
				classFileName.lastIndexOf(".class"));
		} else {
			className = classFileName.substring(sepIndex + 1, 
				classFileName.lastIndexOf(".class"));
		}
		
		//Translation
		PDSStructure PDS = new PDSStructure(classFileName, bits);
		StringBuffer strPDS;
		
		if (newSyntax) {
			strPDS = PDS.printRemopla();
		}
		else {
			strPDS = PDS.print();
		}
		
		//For -D option
		if (dOpt.length() > 0) {
			String[] dVars = dOpt.toString().split("\\s");
			String dVar, dBit;
			int dIndexFrom, dIndexTo, dEqIndex;
			for (int i = 0; i < dVars.length; i++) {
				dEqIndex = dVars[i].indexOf("=");
				if (dEqIndex == -1) {
					System.err.println("Invalid -D format. Usage: -Dclass.field=n, "
						+ "where n is a number\n");
					System.exit(1);
				}
				dVar = dVars[i].substring(0, dEqIndex).replace('.', '_');
				dBit = dVars[i].substring(dEqIndex + 1);
				if ((dIndexFrom = strPDS.indexOf(dVar)) != -1) {
					while (strPDS.charAt(dIndexFrom) != '(')
						dIndexFrom++;
					dIndexTo = ++dIndexFrom;
					while (strPDS.charAt(dIndexTo) != ')')
						dIndexTo++;
					strPDS.replace(dIndexFrom, dIndexTo, dBit);
				}
				else
					System.out.println("Variable " + dVar + " not found, ignored");
			}
		}
		
		//Writes the resulting PDS to a file
		if (outFileName.length() == 0) {
			outFileName = classFileName.replaceAll(".class", ".pds");
		}
		File outFile = new File(outFileName);
		Writer output = null;
		try {
			output = new BufferedWriter(new FileWriter(outFile));
			output.write(strPDS.toString());
		}
		finally {
			if (output != null) output.close();
		}
		
		if (newSyntax) {
			//System.out.println(strPDS);
			System.exit(0);
		}
		
		//Checks execution mode
		String[] mopedCmd = new String[4];
		mopedCmd[0] = "moped";
		mopedCmd[2] = outFileName;
		if (chkAssert) {
			mopedCmd[1] = "-r" + mopedOpt.toString();
			mopedCmd[3] = "q:" + PDSConstant.ASSERTION_ERROR;
		} else if (chkErr) {
			mopedCmd[1] = "-r" + mopedOpt.toString();
			mopedCmd[3] = "q:" + className + "_error_V0";
		} else if (chkLTL) {
			mopedCmd[1] = "-cN" + mopedOpt.toString();
			mopedCmd[3] = "'" + mopedForm + "'";
		} else if (chkReach) {
			mopedCmd[1] = "-r" + mopedOpt.toString();
			mopedCmd[3] = mopedForm;
		} else if (chkTerm) {
			ltlFile = new File(ltlFileName);
			output = null;
			try {
				output = new BufferedWriter(new FileWriter(ltlFile));
				output.write("never {\n");
				output.write("accept_init:\n");
				output.write("T0_init:\n");
				output.write("	if\n");
				output.write("	:: (! ((" + className + "_ok_V0))) -> goto T0_init\n");
				output.write("	fi;\n");
				output.write("}\n");
			} finally {
				if (output != null) output.close();
			}
			mopedCmd[1] = "-cNf" + mopedOpt.toString();
			mopedCmd[3] = ltlFileName;
		}
		
		//Runs moped
		if (chkAssert || chkErr || chkLTL || chkReach || chkTerm)  {
			System.out.println(mopedCmd[0] + " " + mopedCmd[1] + " " 
				+ mopedCmd[2] + " " + mopedCmd[3]);
			Runtime runtime = Runtime.getRuntime();
			StringBuffer returnVal = new StringBuffer();
			String buffer;
			try {
				Process process = runtime.exec(mopedCmd);
				
				BufferedReader in = new BufferedReader(new InputStreamReader(
					process.getInputStream()));
				BufferedReader err = new BufferedReader(new InputStreamReader(
					process.getErrorStream()));
				
				while((buffer = in.readLine()) != null) {
					returnVal.append(buffer).append("\n");
				}
				while((buffer = err.readLine()) != null) {
					returnVal.append(buffer).append("\n");
				}
			} catch (Exception e) {
				System.out.println(e);
			}
			
			//Rewrite output YES/NO
			if (chkAssert) {
				int ansIndex = returnVal.lastIndexOf("YES.");
				
				if (ansIndex != -1) {
					returnVal.replace(ansIndex, ansIndex + 4, 
						"\nRESULT: An assertion is violated.");
				} else {
					ansIndex = returnVal.lastIndexOf("NO.");
					if (ansIndex == -1) {
						System.out.println(returnVal);
						System.exit(1);
					}
					returnVal.replace(ansIndex, ansIndex + 3, 
						"\nRESULT: No assertion is violated.");
				}
			} else if (chkTerm) {
				int ansIndex = returnVal.lastIndexOf("YES.");
				
				if (ansIndex != -1) {
					returnVal.replace(ansIndex, ansIndex + 4, 
						"\nRESULT: The program always terminates.");
				} else {
					ansIndex = returnVal.lastIndexOf("NO.");
					returnVal.replace(ansIndex, ansIndex + 3, 
						"\nRESULT: The program does not always terminate.");
				}
			}
			
			//Handles counterexample
			int startIndex;
			int endIndex;
			CounterExampleHandler cexHandler;
			String cexInput;
		
			if ((startIndex = returnVal.indexOf("--- START ---")) != -1) {
				System.out.println(returnVal.substring(0, startIndex));
				if (chkLTL || chkTerm) {
					endIndex = returnVal.indexOf(
						"Statistics for common resources (overall):");
				} else {
					endIndex = returnVal.indexOf("[ target reached ]");
				}
				cexInput = returnVal.substring(startIndex, endIndex);
				if (javaInput) {
					cexHandler = new CounterExampleHandler(PDS.getClassFileList(), 
							PDS.getPDSClassList(), cexInput, inFileName);
				} else {
					cexHandler = new CounterExampleHandler(PDS.getClassFileList(), 
							PDS.getPDSClassList(), cexInput);
				}
				cexHandler.interpret();
				System.out.println(returnVal.substring(endIndex));
			} else {
				System.out.println(returnVal);
			}
		}

		//Delete LTL file (if exist)
		if (ltlFile != null)
			ltlFile.delete();
	}
}
